import sys
from contextlib import contextmanager

import pytest
from dagster import file_relative_path
from dagster.core.host_representation import (
    ManagedGrpcPythonEnvRepositoryLocationOrigin,
    PipelineHandle,
    RepositoryLocation,
    RepositoryLocationHandle,
)
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin


def get_example_repository_location_handle():
    loadable_target_origin = LoadableTargetOrigin(
        executable_path=sys.executable, python_file=file_relative_path(__file__, "repo.py"),
    )
    location_name = "example_repo_location"

    origin = ManagedGrpcPythonEnvRepositoryLocationOrigin(loadable_target_origin, location_name)

    return RepositoryLocationHandle.create_from_repository_location_origin(origin)


@contextmanager
def get_example_repo_handle():
    with get_example_repository_location_handle() as location_handle:
        yield RepositoryLocation.from_handle(location_handle).get_repository("example_repo").handle


@pytest.fixture
def foo_pipeline_handle():
    with get_example_repo_handle() as repo_handle:
        yield PipelineHandle("foo_pipeline", repo_handle)



import pendulum
from dagster import DagsterInvariantViolationError
from dagster.core.test_utils import instance_for_test
from dagster.daemon.controller import (
    DAEMON_HEARTBEAT_INTERVAL_SECONDS,
    DagsterDaemonController,
    all_daemons_healthy,
    all_daemons_live,
    get_daemon_status,
)
from dagster.utils.error import SerializableErrorInfo


def test_healthy():

    with instance_for_test(
        overrides={
            "run_coordinator": {
                "module": "dagster.core.run_coordinator.queued_run_coordinator",
                "class": "QueuedRunCoordinator",
            },
        }
    ) as instance:
        init_time = pendulum.now("UTC")
        beyond_tolerated_time = init_time.float_timestamp + 100

        controller = DagsterDaemonController(instance)
        assert not all_daemons_healthy(instance, curr_time_seconds=init_time.float_timestamp)
        assert not all_daemons_live(instance, curr_time_seconds=init_time.float_timestamp)

        controller.run_iteration(init_time)
        assert all_daemons_healthy(instance, curr_time_seconds=init_time.float_timestamp)
        assert all_daemons_live(instance, curr_time_seconds=init_time.float_timestamp)

        assert not all_daemons_healthy(instance, curr_time_seconds=beyond_tolerated_time)
        assert not all_daemons_live(instance, curr_time_seconds=beyond_tolerated_time)


def test_healthy_with_different_daemons():
    with instance_for_test() as instance:
        init_time = pendulum.now("UTC")
        controller = DagsterDaemonController(instance)
        controller.run_iteration(init_time)

    with instance_for_test(
        overrides={
            "run_coordinator": {
                "module": "dagster.core.run_coordinator.queued_run_coordinator",
                "class": "QueuedRunCoordinator",
            },
        }
    ) as instance:
        assert not all_daemons_healthy(instance, curr_time_seconds=init_time.float_timestamp)
        assert not all_daemons_live(instance, curr_time_seconds=init_time.float_timestamp)


def test_error_daemon(monkeypatch):
    with instance_for_test(overrides={}) as instance:
        from dagster.daemon.daemon import SensorDaemon

        def run_iteration_error(_):
            raise DagsterInvariantViolationError("foobar")
            yield  # pylint: disable=unreachable

        monkeypatch.setattr(SensorDaemon, "run_iteration", run_iteration_error)
        controller = DagsterDaemonController(instance)
        init_time = pendulum.now("UTC")
        controller.run_iteration(init_time)

        status = get_daemon_status(instance, SensorDaemon.daemon_type(), init_time.float_timestamp)
        assert status.healthy == False
        assert len(status.last_heartbeat.errors) == 1
        assert (
            status.last_heartbeat.errors[0].message.strip()
            == "dagster.core.errors.DagsterInvariantViolationError: foobar"
        )
        assert not all_daemons_healthy(instance, curr_time_seconds=init_time.float_timestamp)
        assert all_daemons_live(instance, curr_time_seconds=init_time.float_timestamp)


def test_multiple_error_daemon(monkeypatch):
    with instance_for_test(overrides={}) as instance:
        from dagster.daemon.daemon import SensorDaemon

        def run_iteration_error(_):
            # ?message stack cls_name cause"
            yield SerializableErrorInfo("foobar", None, None, None)
            yield SerializableErrorInfo("bizbuz", None, None, None)

        monkeypatch.setattr(SensorDaemon, "run_iteration", run_iteration_error)
        controller = DagsterDaemonController(instance)
        init_time = pendulum.now("UTC")
        controller.run_iteration(init_time)

        status = get_daemon_status(instance, SensorDaemon.daemon_type(), init_time.float_timestamp)
        assert status.healthy == False
        assert len(status.last_heartbeat.errors) == 2
        assert status.last_heartbeat.errors[0].message.strip() == "foobar"
        assert status.last_heartbeat.errors[1].message.strip() == "bizbuz"


def test_warn_multiple_daemons(capsys):
    with instance_for_test() as instance:
        init_time = pendulum.now("UTC")
        next_time = init_time.add(seconds=100)

        controller1 = DagsterDaemonController(instance)
        controller1.run_iteration(init_time)
        captured = capsys.readouterr()
        assert "Taking over from another SENSOR daemon process" not in captured.out

        controller2 = DagsterDaemonController(instance)
        controller2.run_iteration(init_time)
        captured = capsys.readouterr()
        assert "Taking over from another SENSOR daemon process" not in captured.out

        controller1.run_iteration(next_time)
        captured = capsys.readouterr()
        assert "Taking over from another SENSOR daemon process" in captured.out



from dagster.core.test_utils import create_run_for_test, poll_for_finished_run
from dagster.utils import merge_dicts
from dagster.utils.external import external_pipeline_from_run

from .utils import setup_instance, start_daemon


def create_run(instance, pipeline_handle, **kwargs):  # pylint: disable=redefined-outer-name
    pipeline_args = merge_dicts(
        {
            "pipeline_name": "foo_pipeline",
            "external_pipeline_origin": pipeline_handle.get_external_origin(),
        },
        kwargs,
    )
    return create_run_for_test(instance, **pipeline_args)


def assert_events_in_order(logs, expected_events):

    logged_events = [log.dagster_event.event_type_value for log in logs if log.is_dagster_event]
    filtered_logged_events = [event for event in logged_events if event in expected_events]

    assert filtered_logged_events == expected_events


def test_queued_runs(tmpdir, foo_pipeline_handle):
    dagster_home_path = tmpdir.strpath
    with setup_instance(
        dagster_home_path,
        """run_coordinator:
    module: dagster.core.run_coordinator
    class: QueuedRunCoordinator
    config:
        dequeue_interval_seconds: 1
    """,
    ) as instance:
        with start_daemon():
            run = create_run(instance, foo_pipeline_handle)
            with external_pipeline_from_run(run) as external_pipeline:
                instance.submit_run(run.run_id, external_pipeline)

                poll_for_finished_run(instance, run.run_id)

                logs = instance.all_logs(run.run_id)
                assert_events_in_order(
                    logs,
                    [
                        "PIPELINE_ENQUEUED",
                        "PIPELINE_DEQUEUED",
                        "PIPELINE_STARTING",
                        "PIPELINE_START",
                        "PIPELINE_SUCCESS",
                    ],
                )



# pylint: disable=protected-access
import os

from dagster import file_relative_path
from dagster.core.instance import DagsterInstance, InstanceRef
from dagster.core.storage.event_log.sql_event_log import (
    SECONDARY_INDEX_ASSET_KEY,
    SqlEventLogStorage,
)
from dagster.utils.test import copy_directory


def test_event_log_asset_key_migration():
    src_dir = file_relative_path(__file__, "snapshot_0_9_22_lazy_asset_index_migration/sqlite")
    with copy_directory(src_dir) as test_dir:
        instance = DagsterInstance.from_ref(
            InstanceRef.from_dir(
                test_dir,
                overrides={
                    "event_log_storage": {
                        "module": "dagster.core.storage.event_log.sqlite.consolidated_sqlite_event_log",
                        "class": "ConsolidatedSqliteEventLogStorage",
                        "config": {"base_dir": os.path.join(test_dir, "history")},
                    }
                },
            )
        )

        # ensure everything is upgraded
        instance.upgrade()

        assert isinstance(instance._event_storage, SqlEventLogStorage)
        assert not instance._event_storage.has_secondary_index(SECONDARY_INDEX_ASSET_KEY)

        old_keys = instance.all_asset_keys()

        assert instance._event_storage.has_secondary_index(SECONDARY_INDEX_ASSET_KEY)

        new_keys = instance.all_asset_keys()

        assert set(old_keys) == set(new_keys)



import datetime
import logging
import re

import pendulum
import pytest
from click.testing import CliRunner
from dagster.core.test_utils import instance_for_test
from dagster.daemon.cli import run_command
from dagster.daemon.controller import DagsterDaemonController
from dagster.daemon.daemon import SchedulerDaemon
from dagster.daemon.run_coordinator.queued_run_coordinator_daemon import QueuedRunCoordinatorDaemon
from dagster.daemon.types import DaemonType


def test_scheduler_instance():
    with instance_for_test(
        overrides={
            "scheduler": {"module": "dagster.core.scheduler", "class": "DagsterDaemonScheduler",},
        }
    ) as instance:
        controller = DagsterDaemonController(instance)

        daemons = controller.daemons

        assert len(daemons) == 2
        assert any(isinstance(daemon, SchedulerDaemon) for daemon in daemons)


def test_run_coordinator_instance():
    with instance_for_test(
        overrides={
            "run_coordinator": {
                "module": "dagster.core.run_coordinator.queued_run_coordinator",
                "class": "QueuedRunCoordinator",
            },
        }
    ) as instance:
        controller = DagsterDaemonController(instance)

        daemons = controller.daemons

        assert len(daemons) == 3
        assert any(isinstance(daemon, QueuedRunCoordinatorDaemon) for daemon in daemons)


def _scheduler_ran(caplog):
    for log_tuple in caplog.record_tuples:
        logger_name, _level, text = log_tuple

        if (
            logger_name == "SchedulerDaemon"
            and "Not checking for any runs since no schedules have been started." in text
        ):
            return True

    return False


def _run_coordinator_ran(caplog):
    for log_tuple in caplog.record_tuples:
        logger_name, _level, text = log_tuple

        if logger_name == "QueuedRunCoordinatorDaemon" and "Poll returned no queued runs." in text:
            return True

    return False


def test_ephemeral_instance():
    runner = CliRunner()
    with pytest.raises(
        Exception,
        match=re.escape(
            "dagster-daemon can't run using an in-memory instance. Make sure the DAGSTER_HOME environment variable has been set correctly and that you have created a dagster.yaml file there."
        ),
    ):
        runner.invoke(run_command, env={"DAGSTER_HOME": ""}, catch_exceptions=False)


def test_different_intervals(caplog):
    with instance_for_test(
        overrides={
            "scheduler": {"module": "dagster.core.scheduler", "class": "DagsterDaemonScheduler",},
            "run_coordinator": {
                "module": "dagster.core.run_coordinator.queued_run_coordinator",
                "class": "QueuedRunCoordinator",
                "config": {"dequeue_interval_seconds": 5},
            },
        }
    ) as instance:
        init_time = pendulum.now("UTC")
        controller = DagsterDaemonController(instance)

        assert caplog.record_tuples == [
            (
                "dagster-daemon",
                logging.INFO,
                "instance is configured with the following daemons: ['QueuedRunCoordinatorDaemon', 'SchedulerDaemon', 'SensorDaemon']",
            )
        ]

        controller.run_iteration(init_time)

        scheduler_daemon = controller.get_daemon(DaemonType.SCHEDULER)
        run_daemon = controller.get_daemon(DaemonType.QUEUED_RUN_COORDINATOR)

        assert scheduler_daemon
        assert (
            controller.get_daemon_last_iteration_time(scheduler_daemon.daemon_type()) == init_time
        )
        assert _scheduler_ran(caplog)

        assert run_daemon
        assert controller.get_daemon_last_iteration_time(run_daemon.daemon_type()) == init_time
        assert _run_coordinator_ran(caplog)
        caplog.clear()

        next_time = init_time + datetime.timedelta(seconds=5)
        controller.run_iteration(next_time)

        # Run coordinator does another iteration, scheduler does not
        assert (
            controller.get_daemon_last_iteration_time(scheduler_daemon.daemon_type()) == init_time
        )
        assert not _scheduler_ran(caplog)

        assert controller.get_daemon_last_iteration_time(run_daemon.daemon_type()) == next_time
        assert _run_coordinator_ran(caplog)
        caplog.clear()

        next_time = init_time + datetime.timedelta(seconds=30)
        controller.run_iteration(next_time)

        # 30 seconds later both daemons do another iteration
        assert (
            controller.get_daemon_last_iteration_time(scheduler_daemon.daemon_type()) == next_time
        )
        assert _scheduler_ran(caplog)

        assert controller.get_daemon_last_iteration_time(run_daemon.daemon_type()) == next_time
        assert _run_coordinator_ran(caplog)



from dagster.daemon.types import DaemonType
from dagster.serdes import deserialize_json_to_dagster_namedtuple
from dagster.utils.error import SerializableErrorInfo


def test_error_backcompat():
    old_heartbeat = '{"__class__": "DaemonHeartbeat", "daemon_id": "foobar", "daemon_type": {"__enum__": "DaemonType.SENSOR"}, "error": {"__class__": "SerializableErrorInfo", "cause": null, "cls_name": null, "message": "fizbuz", "stack": []}, "timestamp": 0.0}'
    heartbeat = deserialize_json_to_dagster_namedtuple(old_heartbeat)
    assert heartbeat.daemon_id == "foobar"
    assert heartbeat.daemon_type == DaemonType.SENSOR
    assert heartbeat.timestamp == 0.0
    assert heartbeat.errors == [SerializableErrorInfo("fizbuz", [], None)]



from dagster import solid
from dagster.core.definitions.decorators.graph import graph
from dagster.core.execution.execute import execute_in_process


def get_solids():
    @solid
    def emit_one(_):
        return 1

    @solid
    def add(_, x, y):
        return x + y

    return emit_one, add


def test_execute_solid():
    emit_one, _ = get_solids()

    result = execute_in_process(emit_one)

    assert result.success


def test_execute_graph():
    emit_one, add = get_solids()

    @graph
    def emit_two():
        return add(emit_one(), emit_one())

    @graph
    def emit_three():
        return add(emit_two(), emit_one())

    result = execute_in_process(emit_three)

    assert result.success
